# SPDX-FileCopyrightText: 2022-present deepset GmbH <info@deepset.ai>
#
# SPDX-License-Identifier: Apache-2.0
from canals.component.component import component
from canals.component.decorators import save_init_params
from canals.component.input_output import ComponentInput, ComponentOutput, VariadicComponentInput
